/* -----------------------------------------------------------------
The VIVA file system - version 2.0

Implemented by Shankar Pasupathy, July 1996. (shankar@pop.uky.edu)
Modified by Vamsi Krishna, July 1997. (vamsi@cslab.uky.edu)
VIVA was developed by Eric Herrin II and Raphael Finkel.
Department of Computer Science,
University of Kentucky, Lexington 40506.
Distribution site: http://www.cslab.uky.edu/~vamsi/viva

This implementation uses code from the Minix filesystem,
(Copyright: Linus Torvalds).

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation. Read the file COPYRIGHT, included with
this distribution, for further details.
--------------------------------------------------------------------*/

#ifndef _LINUX_VIVA_FS_H
#define _LINUX_VIVA_FS_H

#include<linux/types.h>

#define VIVA_ROOT_INO 1
#define VIVA_BLOCK_SIZE 2048
#define VIVA_BLOCK_SIZE_BITS 11
#define VIVA_MAX_INODE_SEGMENTS 64
#define VIVA_MAX_ZONE_SEGMENTS 180
#define VIVA_DIRECT_SMAP_SIZE 219
#define VIVA_INDIRECT_SMAP_SIZE VIVA_BLOCK_SIZE
#define VIVA_SMAP_SIZE (VIVA_DIRECT_SMAP_SIZE+VIVA_BLOCK_SIZE)
#define VIVA_FRAG_SIZE 512
#define VIVA_FRAG_PER_BLOCK (VIVA_BLOCK_SIZE/VIVA_FRAG_SIZE)

/* Not the same as the bogus LINK_MAX in <linux/limits.h>. Oh well. */
#define VIVA_LINK_MAX	250

/* Vamsi: Changed the number of ZMAP Blocks */
#define VIVA_I_MAP_SLOTS	2
#define VIVA_Z_MAP_SLOTS	14
#define VIVA_FS_MAP_SLOTS	VIVA_Z_MAP_SLOTS
#define VIVA_FS_MAP_SIZE 	(VIVA_FS_MAP_SLOTS*VIVA_BLOCK_SIZE)
#define VIVA_FA_MAP_SLOTS	VIVA_FS_MAP_SLOTS
#define VIVA_SUPER_MAGIC	0x2469   	/* original viva fs */
#define VIVA_SUPER_MAGIC2	0x2470		/* viva fs, 30 char names */
#define VIVA_VALID_FS		0x0001		/* Clean fs. */
#define VIVA_ERROR_FS		0x0002		/* fs has errors. */

#define VIVA_INODES_PER_BLOCK ((VIVA_BLOCK_SIZE)/(sizeof (struct viva_inode)))

/* Some useful macros */
#define START_ADDRESS(smap) (*((unsigned long *)(smap)))

/* An entry in the inode sparse bitmap looks like this */
struct viva_smap_entry {
	unsigned long start_addr;
	unsigned char length;
	unsigned char bitmap[0];
};

/*
 * This is the original viva inode layout on disk.
 * Note the 8-bit gid and atime and ctime.
 */
struct viva_inode {
	__u32 i_size;
	__u32 i_time;
	__u32 i_lastblock; 
	__u32 i_frag_block;
	__u16 i_mode;
	__u16 i_uid;
	__u16 i_blocksInUse; 
	__u16 i_gid;
	__u8  i_nlinks;
	__u8  i_frag_no[VIVA_FRAG_PER_BLOCK-1];
	__u8  i_fragsInUse;
	__u8  i_smap[223]; 
	__u32 i_indirect_block;
};
/* 256 byte inode structure */

/*
 * viva super-block data on disk
 */
struct viva_super_block {
	__u16 s_ninodes; /* total number of inodes */
	__u32 s_nzones; /* total number of data zones */
	__u16 s_imap_blocks; /* number of blocks to store inode bitmap */
	__u16 s_zmap_blocks; /* number of blocks to store zone bitmap */
	__u16 s_fsmap_blocks; /* number of blocks to store fragment_sparse bitmap */
	__u16 s_famap_blocks; /* number of blocks to store fragment_alloc bitmap */
	__u16 s_inode_nsegments; /* number of inode segments */
	__u16 s_zone_nsegments; /* number of zone segments */
	__u16 s_inode_seg_size; /* size of all but last inode segment */
	__u16 s_zone_seg_size;
	__u16 s_inode_lastseg_size; /* size of last inode segment */
	__u16 s_zone_lastseg_size;
	__u32 s_firstdatazone; /* where the first data zone begins */
	__u32 s_frag_blocks; /* Number of fragment blocks allocated*/
	__u32 s_max_frag_blocks; /* Max Number of fragment blocks allowed*/
	__u32 s_last_frag_block; /* where the last fragment block is allocated*/
	__u16 s_log_zone_size; /* to modify logical block sizes */
	__u16 s_magic; /* VIVA file system magic number */
	__u16 s_state; /* fs clean or fs has errors */
};


struct viva_dir_entry {
	__u16 inode;
	char name[0];
};

#ifdef __KERNEL__

extern int init_viva_fs(void);

extern int viva_lookup(struct inode * dir,const char * name, int len,
	struct inode ** result);
extern int viva_create(struct inode * dir,const char * name, int len, int mode,
	struct inode ** result);
extern int viva_mkdir(struct inode * dir, const char * name, int len, int mode);
extern int viva_rmdir(struct inode * dir, const char * name, int len);
extern int viva_unlink(struct inode * dir, const char * name, int len);
extern int viva_symlink(struct inode * inode, const char * name, int len,
	const char * symname);
extern int viva_link(struct inode * oldinode, struct inode * dir, const char * name, int len);
extern int viva_mknod(struct inode * dir, const char * name, int len, int mode, int rdev);
extern int viva_rename(struct inode * old_dir, const char * old_name, int old_len,
	struct inode * new_dir, const char * new_name, int new_len, int must_be_dir);
extern struct inode * viva_new_inode(const struct inode * dir);
extern void viva_free_inode(struct inode * inode);
extern unsigned long viva_count_free_inodes(struct super_block *sb);
extern int viva_new_block(struct super_block * sb,unsigned int goal_block, unsigned int blocksInUse);
extern void viva_free_block(struct super_block * sb, int block);
extern unsigned long viva_count_free_blocks(struct super_block *sb);

extern int viva_bmap(unsigned char *, int);
extern int viva_add_hole(struct inode *, int);
extern int viva_break_hole(struct inode *, int, int);
extern int viva_expand_smap(unsigned char *, int, int);
extern int viva_release_smap(unsigned char *, int);

extern long viva_lseek(struct inode *,struct file *,off_t,int);
extern struct buffer_head * viva_getblk(struct inode *, int, int);
extern struct buffer_head * viva_bread(struct inode *, int, int);

extern void viva_truncate(struct inode *);
extern void viva_put_super(struct super_block *);
extern struct super_block *viva_read_super(struct super_block *,void *,int);
extern void viva_write_super(struct super_block *);
extern int viva_remount (struct super_block * sb, int * flags, char * data);
extern void viva_read_inode(struct inode *);
extern void viva_write_inode(struct inode *);
extern void viva_put_inode(struct inode *);
extern void viva_statfs(struct super_block *, struct statfs *, int);
extern int viva_sync_inode(struct inode *);
extern int viva_sync_file(struct inode *, struct file *);
extern void viva_free_frag(struct inode *, int, int);

extern struct inode_operations viva_file_inode_operations;
extern struct inode_operations viva_dir_inode_operations;
extern struct inode_operations viva_symlink_inode_operations;

#endif /* __KERNEL__ */

#endif
