/* -----------------------------------------------------------------
The VIVA file system - version 2.0

Implemented by Shankar Pasupathy, July 1996. (shankar@pop.uky.edu)
Modified by Vamsi Krishna, July 1997. (vamsi@cslab.uky.edu)
VIVA was developed by Eric Herrin II and Raphael Finkel.
Department of Computer Science,
University of Kentucky, Lexington 40506.
Distribution site: http://www.cslab.uky.edu/~vamsi/viva

This implementation uses code from the Minix filesystem,
(Copyright: Linus Torvalds).

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation. Read the file COPYRIGHT, included with
this distribution, for further details.
--------------------------------------------------------------------*/

/* bitmap.c contains the code that handles the inode and block bitmaps */

/*
#ifdef MODULE
#include <linux/module.h>
#endif
*/

#include <linux/sched.h>
#include <linux/viva_fs.h>
#include <linux/stat.h>
#include <linux/kernel.h>
#include <linux/string.h>
#include <linux/locks.h>

#include <asm/bitops.h>

/* Include the code to modify segment counts */
#include "segment_counts.c"

static int nibblemap[] = { 0,1,1,2,1,2,2,3,1,2,2,3,2,3,3,4 };

/*------------------------------------------------------------------------
count_used: counts the number of 1s in a given bitmap. Called from statfs,
which in turn is called by the utility program df.
--------------------------------------------------------------------------*/
static unsigned long count_used(struct buffer_head *map[], unsigned numblocks,
	unsigned numbits)
{
	unsigned count, temp, end, sum = 0;
	struct buffer_head *bh;
  
	for (count=0; (count<numblocks) && numbits; count++) {
		if (!(bh=map[count])) 
			return(0);
		if (numbits >= (8*VIVA_BLOCK_SIZE)) { 
			end = VIVA_BLOCK_SIZE;
			numbits -= 8*VIVA_BLOCK_SIZE;
		} else {
			int tmp;
			end = numbits >> 3;
			numbits &= 0x7;
			tmp = bh->b_data[end] & ((1<<numbits)-1);
			sum += nibblemap[tmp&0xf] + nibblemap[(tmp>>4)&0xf];
			numbits = 0;
		}  
		for (temp=0; temp<end; temp++)
			sum += nibblemap[bh->b_data[temp] & 0xf] 
				+ nibblemap[(bh->b_data[temp]>>4)&0xf];
	}
	return(sum);
}


void viva_free_block(struct super_block * sb, int block)
{
	struct buffer_head * bh;
	unsigned int bit,zone,zone_slot,seg;
	unsigned char modified_byte;

	if (!sb) {
		printk("trying to free block on nonexistent device\n");
		return;
	}
	lock_super(sb);
	if (block < sb->u.viva_sb.s_firstdatazone ||
	    block >= sb->u.viva_sb.s_nzones) {
		printk("viva_free_block: Trying to free block %d not in datazone\n",block);
		unlock_super(sb);
		return;
	}
	bh = get_hash_table(sb->s_dev,block,VIVA_BLOCK_SIZE);
	if (bh)
		clear_bit(BH_Dirty, &bh->b_state);
	brelse(bh);
	zone = block - sb->u.viva_sb.s_firstdatazone;
	bit = zone & ((VIVA_BLOCK_SIZE << 3) - 1);
	/* a & b is the same as a % (b+1) */
		
	zone_slot = zone>>(VIVA_BLOCK_SIZE_BITS + 3);
	bh = sb->u.viva_sb.s_zmap[zone_slot];
	if (!bh) {
		printk("viva_free_block: nonexistent bitmap buffer\n");
		unlock_super(sb);
		return;
	}
	if (!clear_bit(bit,bh->b_data)) {
		printk("free_block (%04x:%d): bit already cleared\n",sb->s_dev,block);
	}
	else {
		/* important : change segment counts only after bitmaps have been
		modified */

		/* bit ranges from 0 to 8191 at this point */
		modified_byte = bh->b_data[bit/8];
		seg = zone / (sb->u.viva_sb.s_zone_seg_size*8);
		modify_zone_segment_counts(sb, seg, modified_byte, bit % 8, 1);
	}
	unlock_super(sb);
	mark_buffer_dirty(bh, 1);
	return;
}

	
int viva_new_block(struct super_block * sb, unsigned int goal,unsigned int blocksInUse)
{
	struct buffer_head * bh;
	unsigned int new_block,seg,seg_count,seg_length;
	unsigned int zmap_slot, zmap_byte, zmap_bit;
	unsigned char old_byte;
	int real_boundary, boundary, search_count,search_length;
	 /* don't make boundary an unsigned int !! */

	if (!sb) {
		printk("VIVA: trying to get new block from nonexistent device\n");
		return 0;
	}

	lock_super(sb); 

	/*
	First check to see that the goal lies within proper bounds 
	 */
	if (goal >= sb->u.viva_sb.s_nzones) {
		printk("Goal exceeds bounds\n"); 
		goal = sb->u.viva_sb.s_firstdatazone;
	}
	/* Reduce the goal_block down to a 0-... range */
	goal = goal - sb->u.viva_sb.s_firstdatazone;

	/* 
	First check if the goal is free.
	The goal lies in a zone bitmap slot (there are 8 slots),
	zmap_slot = goal >> 13	or goal/8192
	and in the byte, zmap_byte =(goal/8) - (zmap_slot * 1024) within
	the slot
	 */

	zmap_slot = goal/(VIVA_BLOCK_SIZE << 3);
	zmap_byte = (goal/8) - (zmap_slot * VIVA_BLOCK_SIZE);
	zmap_bit = goal % 8;

	/* load the zone bitmap corresponding to zmap_slot */
	bh=sb->u.viva_sb.s_zmap[zmap_slot];

	/* find which segment this goal lies in */
	seg = goal / (sb->u.viva_sb.s_zone_seg_size * 8);

	old_byte = bh->b_data[zmap_byte];
	if (!test_bit(zmap_bit, bh->b_data + zmap_byte)) {
		/* the goal is free ! */
		/* change segment counts to take into account the allocation of a new block */
		modify_zone_segment_counts(sb, seg, old_byte, zmap_bit, -1);
		new_block = sb->u.viva_sb.s_firstdatazone + goal;
		set_bit(zmap_bit,bh->b_data+zmap_byte);
		goto got_block;
	}	

	/*
	The goal bit was not free.
	Now check if we can find a free bit within 32 bits of the goal
	*/
	search_count = 7-zmap_bit;
	/* at this stage, zmap_bit is between 0 and 7 */
	/* search_count is used to search within 32 bits from the goal */
	if ((zmap_bit = find_next_zero_bit(bh->b_data+zmap_byte,8,zmap_bit)) < 8)
			goto got_block_within_32;

	while ((search_count += 8) < 32) {
		zmap_byte++;
		if (zmap_byte > ((VIVA_BLOCK_SIZE)-1)) {
		/* we are crossing a bitmap slot boundary */
			zmap_byte=0;
			zmap_slot++;
			bh = sb->u.viva_sb.s_zmap[zmap_slot];
			zmap_bit = 0;
		}
		if ((zmap_bit = find_first_zero_bit(bh->b_data+zmap_byte,8)) < 8)
				goto got_block_within_32;
	}	
	/* Too bad - we couldn't find a free bit within 32 bits of the goal ! */
	goto use_boundary_method;

got_block_within_32:
		/*
		Excellent - we found a free bit within 32 bits of the goal. 
		*/
		/* Always compute old_byte before we set the bit in the bitmap */
		old_byte = bh->b_data[zmap_byte + (zmap_bit/8)];
		set_bit(zmap_bit,bh->b_data+zmap_byte);
		/* Figure out the block number */
		new_block = sb->u.viva_sb.s_firstdatazone + (zmap_slot * (VIVA_BLOCK_SIZE << 3)) + (zmap_byte * 8) + zmap_bit;
		/* recompute the segment number */
		seg = (new_block-sb->u.viva_sb.s_firstdatazone) / (sb->u.viva_sb.s_zone_seg_size * 8);
		modify_zone_segment_counts(sb, seg, old_byte, zmap_bit % 8,-1);
		goto got_block;

use_boundary_method:
	/* now use the blocksInUse by the file which requested a new block,
           to decide what bit boundary to use */
	switch(blocksInUse) {
		case 0:	boundary = 0; /* ONE_BIT_BOUNDARY */
			break;
		case 1: boundary = 1; /* TWO_BIT_BOUNDARY */
			break;
		case 2:
		case 3:
		case 4: boundary = 2; /* FOUR_BIT_BOUNDARY */
			break;
		default:boundary = 3; /*EIGHT_BIT_BOUNDARY */
	}

	/*	
	Search for a free bit on requested boundary, and then on lower
	boundaries.
	There's a redundant check here since we don't update seg to take
	into account a failed search within 32 bits of the goal.
	However, it is cheaper to just search the segments rather than 
	recompute the segment number here.
	*/

	while (boundary > -1) {
		/* look at appropriate segments */
		for (seg_count = seg; seg_count < sb->u.viva_sb.s_zone_nsegments; seg_count++) {
			if (sb->u.viva_sb.s_free_zones[boundary][seg_count]) {
				goto got_free_boundary;
			}
		}
		/* cycle through segments backward */
		for (seg_count = 0; seg_count < seg ; seg_count++) {
			if (sb->u.viva_sb.s_free_zones[boundary][seg_count]) {
				goto got_free_boundary;
			}
		}
		/*If we get here, we didn't succeed in finding a free bit on
		the requested boundary. So cycle through smaller boundaries.*/

		boundary = boundary - 1;
	}
	/* If we got here, we are out of luck. There are no free blocks 
	available */
	printk("There are no free blocks available\n");
	unlock_super(sb);
	return -1;

got_free_boundary:
	/* seg_count tells us which segment has a free bit */
	/* The variable `boundary' contains the bit boundary */
	/* Need to get the appropriate bitmap using seg_count */

	goal = seg_count * sb->u.viva_sb.s_zone_seg_size;
	zmap_slot = goal/VIVA_BLOCK_SIZE;
	zmap_byte = (goal) - (zmap_slot * VIVA_BLOCK_SIZE);
	seg_length = (seg_count-sb->u.viva_sb.s_zone_nsegments+1)?
		sb->u.viva_sb.s_zone_seg_size:sb->u.viva_sb.s_zone_lastseg_size;
	/* fetch the appropriate zone bitmap */
	bh=sb->u.viva_sb.s_zmap[zmap_slot];
	/*	
	zmap_byte is the byte at which the goal segment starts.
	There's a horrible problem when the segment straddles 
	a bitmap slot boundary. This will happen when 
	zmap_byte + seg_length > 8192
	Vamsi: 1024 not 8192 ???.
	*/		
		
	if ((zmap_byte + seg_length) > (VIVA_BLOCK_SIZE)-1)
		search_length = (VIVA_BLOCK_SIZE)-zmap_byte;
	else 
		search_length = seg_length;

start_boundary_search:
	zmap_bit = find_first_zero_bit(bh->b_data + zmap_byte,search_length*8);
	if (zmap_bit >= (search_length*8)) {
		/* maybe we are crossing a zone bitmap slot */
		if ((zmap_byte + seg_length) > ((VIVA_BLOCK_SIZE)-1)) {
			/* need to load another part of the zone bitmap */
			zmap_slot++;
			zmap_byte = 0;
			search_length = seg_length - search_length;
		   	bh=sb->u.viva_sb.s_zmap[zmap_slot];
			goto start_boundary_search;
		}
		else {
			printk("Vamsi: viva_new_block possible error 1 \n");
			printk("Vamsi: zmap_byte = %d %d, search_length = %d (%d %d) boundary = %d (%d) \n",zmap_byte, zmap_bit, search_length, seg_length,seg_count,boundary,blocksInUse);
			unlock_super(sb);
			return -1;
		}
	}
	/* Nasty kludge follows */
	real_boundary = 1<<boundary;

	while (((zmap_bit % real_boundary) != 0) && (zmap_bit < (search_length*8))) 
	{
		zmap_bit = find_next_zero_bit(bh->b_data + zmap_byte, search_length*8, zmap_bit + 1);
		if (zmap_bit >= (search_length*8)) {
			if ((zmap_byte + seg_length) > ((VIVA_BLOCK_SIZE)-1)) {
				/* need to load another part of the zone bitmap */	
				zmap_slot++;
				zmap_byte = 0;
				search_length = seg_length - search_length;
		   		bh=sb->u.viva_sb.s_zmap[zmap_slot];
				goto start_boundary_search;
			}
			else {
				printk("Vamsi: viva_new_block possible error 2 \n");
				unlock_super(sb);
				return -1;
			}
		}		
	}
	/* Now we have found a free bit on a desired boundary */
	/* Remember that zmap_bit is in the range 0..search_length*8 - 1*/
	old_byte = bh->b_data[zmap_byte + zmap_bit/8];
	set_bit(zmap_bit,bh->b_data+zmap_byte);
        modify_zone_segment_counts(sb, seg_count, old_byte, zmap_bit % 8, -1);

	/* ALWAYS set the bit before modifying segment counts or else
	we won't be able to recover after a filesystem crash */
	new_block = sb->u.viva_sb.s_firstdatazone + (zmap_slot*(VIVA_BLOCK_SIZE << 3)) + (zmap_byte*8) + zmap_bit;

got_block:	
	if (new_block > (sb->u.viva_sb.s_nzones - 1)) {
		printk("VIVA: File system is full\n");
		return -1;
	}
	mark_buffer_uptodate(bh,1); 
	mark_buffer_dirty(bh, 1);
	/* now check if we can get that block */
	if (!(bh = getblk(sb->s_dev,new_block,VIVA_BLOCK_SIZE))) {
		printk("VIVA: new_block: cannot get block\n");
		unlock_super(sb);
		return(-1);
	}
	memset(bh->b_data, 0, VIVA_BLOCK_SIZE); /* zero-out all entries */
	mark_buffer_uptodate(bh,1); 
	mark_buffer_dirty(bh, 1);
	brelse(bh);
	unlock_super(sb);
	return new_block;
}

unsigned long viva_count_free_blocks(struct super_block *sb)
{
#if 0
	/* this code won't work. Will work only if zones - firstdatazone
	 is a multiple of eight, due to a very complicated problem in mkfs.
	See my report for the reason why this method won't work */	
	unsigned long count_free_blocks = 0;
	for (i=0;i<sb->u.viva_sb.s_zone_nsegments;i++) 
		count_free_blocks += sb->u.viva_sb.s_free_zones[0][i];
	return count_free_blocks;
#endif
#if 1
	return (sb->u.viva_sb.s_nzones - count_used(sb->u.viva_sb.s_zmap,sb->u.viva_sb.s_zmap_blocks,sb->u.viva_sb.s_nzones))
		 << sb->u.viva_sb.s_log_zone_size;
#endif
}

void viva_free_inode(struct inode * inode)
{
	struct buffer_head * bh;
	unsigned long ino;

	if (!inode)
		return;
	if (!inode->i_dev) {
		printk("free_inode: inode has no device\n");
		return;
	}
	if (inode->i_count != 1) {
		printk("free_inode: inode has count=%d\n",inode->i_count);
		return;
	}
	if (inode->i_nlink) {
		printk("free_inode: inode has nlink=%d\n",inode->i_nlink);
		return;
	}
	if (!inode->i_sb) {
		printk("free_inode: inode on nonexistent device\n");
		return;
	}
	if (inode->i_ino < 1 || inode->i_ino >= inode->i_sb->u.viva_sb.s_ninodes) {
		printk("free_inode: inode 0 or nonexistent inode\n");
		return;
	}
	ino = inode->i_ino;
	if (inode->u.viva_i.i_cached_indirect_block) {
		/* release indirect block1 held in memory */
		if (inode->u.viva_i.indirect_smap_bh) {
			mark_buffer_dirty(inode->u.viva_i.indirect_smap_bh,1);
			brelse(inode->u.viva_i.indirect_smap_bh);
		}
		/* mark the block as free */
		viva_free_block(inode->i_sb,inode->u.viva_i.i_cached_indirect_block);
	}	
	if (!(bh=inode->i_sb->u.viva_sb.s_imap[ino >> (VIVA_BLOCK_SIZE_BITS+3)])) {
		printk("free_inode: nonexistent imap in superblock\n");
		return;
	}
	clear_inode(inode);
	if (!clear_bit(ino & ((VIVA_BLOCK_SIZE << 3)-1), bh->b_data))
		printk("free_inode: bit %lu already cleared.\n",ino);
	mark_buffer_dirty(bh, 1);
}

struct inode * viva_new_inode(const struct inode * dir)
{
	struct super_block * sb;
	struct inode * inode;
	struct buffer_head * bh=NULL;
	int imap_slots,bit_offset,num_frags;

	if (!dir || !(inode = get_empty_inode()))
		return NULL;
	sb = dir->i_sb;
	inode->i_sb = sb;
	inode->i_flags = inode->i_sb->s_flags;
	bit_offset = VIVA_BLOCK_SIZE << 3;
	for (imap_slots=0 ; imap_slots<inode->i_sb->u.viva_sb.s_imap_blocks ; imap_slots++)
		if ((bh = inode->i_sb->u.viva_sb.s_imap[imap_slots]) != NULL)
			if((bit_offset=find_first_zero_bit(bh->b_data, VIVA_BLOCK_SIZE<<3))
					< (VIVA_BLOCK_SIZE << 3))
				break;
	if (!bh || bit_offset >= (VIVA_BLOCK_SIZE << 3)) {
		iput(inode);
		return NULL;
	}
	if (set_bit(bit_offset,bh->b_data)) {	/* shouldn't happen */
		printk("new_inode: bit already set");
		iput(inode);
		return NULL;
	}
	mark_buffer_dirty(bh, 1);
	bit_offset += imap_slots*(VIVA_BLOCK_SIZE << 3);
	if (!bit_offset || bit_offset >= inode->i_sb->u.viva_sb.s_ninodes) {
		iput(inode);
		return NULL;
	}
	inode->i_count = 1;
	inode->i_nlink = 1;
	inode->i_dev = sb->s_dev;
	inode->i_uid = current->fsuid;
	inode->i_gid = (dir->i_mode & S_ISGID) ? dir->i_gid : current->fsgid;
	inode->u.viva_i.i_blocksInUse = 0;
	inode->u.viva_i.i_lastblock = 0;
	memset(inode->u.viva_i.i_cached_smap,0,VIVA_SMAP_SIZE);
	inode->u.viva_i.i_cached_indirect_block = 0;
	inode->u.viva_i.indirect_smap_bh = NULL;
	inode->u.viva_i.i_iblock_dirty = 1;
	inode->u.viva_i.i_fragsInUse = 0;
	inode->u.viva_i.i_frag_block = 0;
	for(num_frags=0; num_frags < VIVA_FRAG_PER_BLOCK-1; num_frags++)
		inode->u.viva_i.i_frag_no[num_frags] = 0;
	inode->i_dirt = 1;
	inode->i_ino = bit_offset;
	inode->i_mtime = inode->i_atime = inode->i_ctime = CURRENT_TIME;
	inode->i_op = NULL;
	inode->i_blocks = inode->i_blksize = 0;
	insert_inode_hash(inode);
	return inode;
}

unsigned long viva_count_free_inodes(struct super_block *sb)
{
	return sb->u.viva_sb.s_ninodes - count_used(sb->u.viva_sb.s_imap,sb->u.viva_sb.s_imap_blocks,sb->u.viva_sb.s_ninodes);
}
